
.wy-nav-side {
    padding-bottom: 3em;
}

table.docutils {
    width: 100%;
    table-layout: fixed;
}

.wy-table td, .rst-content table.docutils td {
    overflow: hidden;
}
/*
table.docutils span.scrollable {
    display: block;
    overflow: auto;
}
*/

table.rows p {
    font-size: 100%;
}

table.rows th {
    background-color: #2980B9;
    border-style: solid solid solid solid;
    border-width: 0px 0px 0px 0px;
    border-color: #AAAAAA;
    text-align: center;
    color: white;
}
table.rows td {
    white-space: normal;
    border-style: solid solid solid solid;
    border-width: 0px 0px 0px 0px;
    border-color: #AAAAAA;
}

table.rows tr {
    border-style: solid solid solid solid;
    border-width: 0px 0px 0px 0px;
    border-color: #AAAAAA;
}

table.rows tr:nth-child(even) {
    background-color: white;
}
table.rows tr:nth-child(odd) {
    background-color: grey;
}

div.nboutput div.output_area {
    max-height: 400px;
}

.highlight {
    background: #f5f5f5;
}

.rst-content code.literal {
    background-color: #f5f5f5;
    color: #404040;
}


.rst-content code.download {
    color: rgb(41, 128, 185);
    background-color: white;
}

/* webgui preview image (overlay and message "Click to load interactive webgui" on hover) */
.webgui_container img {
    width: 100%;
    height: 100%;
}

.webgui_container {
    position: relative;
}

.webgui_overlay {
    position: absolute;
    transition: all 0.3s ease;
    opacity: 0;

    height: 100%;
    width: 100%;
    top: 0;
    background-color: #808080;
}

.webgui_container:hover .webgui_overlay {
    opacity: 0.7;
}

.webgui_tooltip .webgui_tooltiptext {
    visibility: hidden;
    background-color: black;
    color: #fff;
    padding: 5px 10px;
    left: 50%;
    border-radius: 6px;
    position: absolute;
    transform: translate(-50%, 10%);
    z-index: 1;
}

.webgui_tooltip:hover .webgui_tooltiptext {
    visibility: visible;
}

.dg .cr {
    margin-left: 0;
}

.dg li.folder {
  list-style: none;
    margin-left: 0;
}

.dg .closed li:not(.title), .dg .closed ul li, .dg .closed ul li>* {
    list-style: none;
    overflow: hidden;
    border: 0;
    margin-left: 0;
}

.folder .dg {
    margin-bottom: 0;

}
